from datetime import datetime
from typing import Annotated, Literal, Sequence
from pydantic import AliasChoices, BaseModel, Field, JsonValue
from .base import *


class StreamPartTextStart(BaseModel):
    type: Literal["text-start"] = "text-start"
    id: str
    provider_metadata: ProviderMetadata | None = Field(
        default=None,
        validation_alias=AliasChoices("providerMetadata", "provider_metadata"),
        serialization_alias="providerMetadata",
    )


class StreamPartTextDelta(BaseModel):
    type: Literal["text-delta"] = "text-delta"
    delta: str
    id: str
    provider_metadata: ProviderMetadata | None = Field(
        default=None,
        validation_alias=AliasChoices("providerMetadata", "provider_metadata"),
        serialization_alias="providerMetadata",
    )


class StreamPartTextEnd(BaseModel):
    type: Literal["text-end"] = "text-end"
    id: str
    provider_metadata: ProviderMetadata | None = Field(
        default=None,
        validation_alias=AliasChoices("providerMetadata", "provider_metadata"),
        serialization_alias="providerMetadata",
    )


class StreamPartReasoningStart(BaseModel):
    type: Literal["reasoning-start"] = "reasoning-start"
    id: str
    provider_metadata: ProviderMetadata | None = Field(
        default=None,
        validation_alias=AliasChoices("providerMetadata", "provider_metadata"),
        serialization_alias="providerMetadata",
    )


class StreamPartReasoningDelta(BaseModel):
    type: Literal["reasoning-delta"] = "reasoning-delta"
    delta: str
    id: str
    provider_metadata: ProviderMetadata | None = Field(
        default=None,
        validation_alias=AliasChoices("providerMetadata", "provider_metadata"),
        serialization_alias="providerMetadata",
    )


class StreamPartReasoningEnd(BaseModel):
    type: Literal["reasoning-end"] = "reasoning-end"
    id: str
    provider_metadata: ProviderMetadata | None = Field(
        default=None,
        validation_alias=AliasChoices("providerMetadata", "provider_metadata"),
        serialization_alias="providerMetadata",
    )


class StreamPartToolInputStart(BaseModel):
    type: Literal["tool-input-start"] = "tool-input-start"
    tool_name: str = Field(
        validation_alias=AliasChoices("toolName", "tool_name"),
        serialization_alias="toolName",
    )
    provider_executed: bool | None = Field(
        default=None,
        validation_alias=AliasChoices("providerExecuted", "provider_executed"),
        serialization_alias="providerExecuted",
    )
    id: str
    provider_metadata: ProviderMetadata | None = Field(
        default=None,
        validation_alias=AliasChoices("providerMetadata", "provider_metadata"),
        serialization_alias="providerMetadata",
    )


class StreamPartToolInputDelta(BaseModel):
    type: Literal["tool-input-delta"] = "tool-input-delta"
    delta: str
    id: str
    provider_metadata: ProviderMetadata | None = Field(
        default=None,
        validation_alias=AliasChoices("providerMetadata", "provider_metadata"),
        serialization_alias="providerMetadata",
    )


class StreamPartToolInputEnd(BaseModel):
    type: Literal["tool-input-end"] = "tool-input-end"
    id: str
    provider_metadata: ProviderMetadata | None = Field(
        default=None,
        validation_alias=AliasChoices("providerMetadata", "provider_metadata"),
        serialization_alias="providerMetadata",
    )


class StreamPartStreamStart(BaseModel):
    type: Literal["stream-start"] = "stream-start"
    warnings: Sequence[Warning]


class StreamPartResponseMetadata(ResponseMetadata):
    type: Literal["response-metadata"] = "response-metadata"


class StreamPartFinish(BaseModel):
    type: Literal["finish"] = "finish"
    usage: Usage
    finish_reason: FinishReason = Field(
        validation_alias=AliasChoices("finishReason", "finish_reason"),
        serialization_alias="finishReason",
    )


class StreamPartRaw(BaseModel):
    type: Literal["raw"] = "raw"
    raw_value: JsonValue = Field(
        validation_alias=AliasChoices("rawValue", "raw_value"),
        serialization_alias="rawValue",
    )


class StreamPartError(BaseModel):
    type: Literal["error"] = "error"
    error: JsonValue


type StreamPart = Annotated[
    StreamPartTextStart
    | StreamPartTextDelta
    | StreamPartTextEnd
    | StreamPartReasoningStart
    | StreamPartReasoningDelta
    | StreamPartReasoningEnd
    | StreamPartToolInputStart
    | StreamPartToolInputDelta
    | StreamPartToolInputEnd
    | ToolCall
    | ToolResult
    | File
    | Source
    | StreamPartStreamStart
    | StreamPartResponseMetadata
    | StreamPartFinish
    | StreamPartRaw
    | StreamPartError,
    Field(discriminator="type"),
]



from agentia.spec import *


class History:
    def __init__(self) -> None:
        self.__instructions: list[str] = []
        self.__non_instruction_messages: list[
            UserMessage | AssistantMessage | ToolMessage
        ] = []

    def add_instructions(self, instruction: str) -> None:
        self.__instructions.append(instruction)

    def __get_instructions(self) -> str:
        return "\n\n".join(self.__instructions)

    def add(self, *messages: UserMessage | AssistantMessage | ToolMessage) -> None:
        self.__non_instruction_messages.extend(messages)

    def get(self, include_instructions: bool = True) -> list[Message]:
        messages: list[Message] = []
        if include_instructions:
            if i := self.__get_instructions():
                messages.append(SystemMessage(content=i))
        messages.extend(self.__non_instruction_messages)
        return messages

    def clear(self) -> None:
        self.__non_instruction_messages = []

    def load(
        self, messages: Sequence[Message], exclude_instructions: bool = False
    ) -> None:
        self.clear()
        for m in messages:
            if isinstance(m, SystemMessage):
                if not exclude_instructions:
                    self.__instructions.append(m.content)
            else:
                self.__non_instruction_messages.append(m)


__all__ = ["History"]



from enum import Enum
import inspect
from typing import Type
from .base import *
from openai.lib._parsing._completions import to_strict_json_schema  # type: ignore


class MessageBase(BaseModel): ...


class SystemMessage(MessageBase):
    role: Literal["system"] = "system"
    content: str
    provider_options: ProviderOptions | None = Field(
        default=None,
        validation_alias=AliasChoices("providerOptions", "providerOptions"),
        serialization_alias="providerOptions",
    )


class MessagePartBase(BaseModel): ...


class MessagePartText(MessagePartBase):
    type: Literal["text"] = "text"
    text: str
    provider_options: ProviderOptions | None = Field(
        default=None,
        validation_alias=AliasChoices("providerOptions", "provider_options"),
        serialization_alias="providerOptions",
    )


class MessagePartReasoning(MessagePartBase):
    type: Literal["reasoning"] = "reasoning"
    text: str
    provider_options: ProviderOptions | None = Field(
        default=None,
        validation_alias=AliasChoices("providerOptions", "provider_options"),
        serialization_alias="providerOptions",
    )


class MessagePartFile(MessagePartBase):
    type: Literal["file"] = "file"
    data: DataContent
    media_type: str = Field(
        validation_alias=AliasChoices("mediaType", "media_type"),
        serialization_alias="mediaType",
    )
    filename: str | None = None
    provider_options: ProviderOptions | None = Field(
        default=None,
        validation_alias=AliasChoices("providerOptions", "provider_options"),
        serialization_alias="providerOptions",
    )


class MessagePartToolCall(MessagePartBase):
    type: Literal["tool-call"] = "tool-call"
    tool_call_id: str = Field(
        validation_alias=AliasChoices("toolCallId", "tool_call_id"),
        serialization_alias="toolCallId",
    )
    tool_name: str = Field(
        validation_alias=AliasChoices("toolName", "tool_name"),
        serialization_alias="toolName",
    )
    input: JsonValue
    provider_executed: bool | None = Field(
        default=None,
        validation_alias=AliasChoices("providerExecuted", "provider_executed"),
        serialization_alias="providerExecuted",
    )
    provider_options: ProviderOptions | None = Field(
        default=None,
        validation_alias=AliasChoices("providerOptions", "provider_options"),
        serialization_alias="providerOptions",
    )


class ToolResultOutputText(BaseModel):
    type: Literal["text"] = "text"
    value: str


class ToolResultOutputJson(BaseModel):
    type: Literal["json"] = "json"
    value: JsonValue


class ToolResultOutputExecutionDenied(BaseModel):
    type: Literal["execution-denied"] = "execution-denied"
    reason: str | None = None


class ToolResultOutputErrorText(BaseModel):
    type: Literal["error-text"] = "error-text"
    value: str


class ToolResultOutputErrorJson(BaseModel):
    type: Literal["error-json"] = "error-json"
    value: JsonValue


class ToolResultOutputContentPartText(BaseModel):
    type: Literal["text"] = "text"
    text: str


class ToolResultOutputContentPartMedia(BaseModel):
    type: Literal["media"] = "media"
    data: str
    media_type: str = Field(
        validation_alias=AliasChoices("mediaType", "media_type"),
        serialization_alias="mediaType",
    )


type ToolResultOutputContentPart = Annotated[
    ToolResultOutputContentPartText | ToolResultOutputContentPartMedia,
    Field(discriminator="type"),
]


class ToolResultOutputContent(BaseModel):
    type: Literal["content"] = "content"
    value: Sequence[ToolResultOutputContentPart]


type ToolResultOutput = Annotated[
    ToolResultOutputText
    | ToolResultOutputJson
    | ToolResultOutputExecutionDenied
    | ToolResultOutputErrorText
    | ToolResultOutputErrorJson
    | ToolResultOutputContent,
    Field(discriminator="type"),
]


class MessagePartToolResult(MessagePartBase):
    type: Literal["tool-result"] = "tool-result"
    tool_call_id: str = Field(
        validation_alias=AliasChoices("toolCallId", "tool_call_id"),
        serialization_alias="toolCallId",
    )
    tool_name: str = Field(
        validation_alias=AliasChoices("toolName", "tool_name"),
        serialization_alias="toolName",
    )
    output: ToolResultOutput
    provider_options: ProviderOptions | None = Field(
        default=None,
        validation_alias=AliasChoices("providerOptions", "provider_options"),
        serialization_alias="providerOptions",
    )


type MessagePart = Annotated[
    MessagePartText
    | MessagePartReasoning
    | MessagePartFile
    | MessagePartToolCall
    | MessagePartToolResult,
    Field(discriminator="type"),
]

type UserMessagePart = Annotated[
    MessagePartText | MessagePartFile,
    Field(discriminator="type"),
]

type AssistantMessagePart = Annotated[
    MessagePartText
    | MessagePartReasoning
    | MessagePartFile
    | MessagePartToolCall
    | MessagePartToolResult,
    Field(discriminator="type"),
]


class UserMessage(MessageBase):
    role: Literal["user"] = "user"
    content: Sequence[UserMessagePart]
    provider_options: ProviderOptions | None = Field(
        default=None,
        validation_alias=AliasChoices("providerOptions", "providerOptions"),
        serialization_alias="providerOptions",
    )

    @property
    def text(self) -> str:
        text = ""
        for part in self.content:
            if isinstance(part, MessagePartText):
                text += part.text
        return text


type ObjectType = BaseModel | Enum | str | int | float | bool | None | list[
    "ObjectType"
] | dict[str, "ObjectType"] | tuple["ObjectType", ...]


class _Result[X](BaseModel):
    result: X = Field(..., description="The result", title="Result")


class AssistantMessage(MessageBase):
    role: Literal["assistant"] = "assistant"
    content: Sequence[AssistantMessagePart]
    provider_options: ProviderOptions | None = Field(
        default=None,
        validation_alias=AliasChoices("providerOptions", "providerOptions"),
        serialization_alias="providerOptions",
    )

    @property
    def reasoning(self) -> str | None:
        text = ""
        for part in self.content:
            if isinstance(part, MessagePartReasoning):
                text += part.text
        return text

    @property
    def text(self) -> str:
        text = ""
        for part in self.content:
            if isinstance(part, MessagePartText):
                text += part.text
        return text

    def parse[T: ObjectType](self, return_type: type[T]) -> T:
        class Result(_Result[return_type]): ...

        if inspect.isclass(return_type) and issubclass(return_type, BaseModel):
            response_format = return_type
        else:
            response_format = Result

        json_string = ""
        for part in self.content:
            if isinstance(part, MessagePartText):
                json_string += part.text

        result = response_format.model_validate_json(json_string)
        if isinstance(result, _Result):
            return result.result
        else:
            return result

    @staticmethod
    def from_contents(contents: Sequence[Content]) -> "AssistantMessage":
        parts = []
        for c in contents:
            if c.type == "text":
                parts.append(MessagePartText(text=c.text))
            elif c.type == "reasoning":
                parts.append(MessagePartReasoning(text=c.text))
            elif c.type == "file":
                parts.append(MessagePartFile(data=c.data, media_type=c.media_type))
            elif c.type == "tool-call":
                parts.append(
                    MessagePartToolCall(
                        tool_call_id=c.tool_call_id,
                        tool_name=c.tool_name,
                        input=c.input,
                        provider_executed=c.provider_executed,
                    )
                )
            else:
                raise Exception(f"Unknown content type: {c.type}")
        assert len(parts) > 0
        return AssistantMessage(content=parts)


class ToolMessage(MessageBase):
    role: Literal["tool"] = "tool"
    content: Sequence[MessagePartToolResult]
    provider_options: ProviderOptions | None = Field(
        default=None,
        validation_alias=AliasChoices("providerOptions", "providerOptions"),
        serialization_alias="providerOptions",
    )


type Message = Annotated[
    SystemMessage | UserMessage | AssistantMessage | ToolMessage,
    Field(discriminator="role"),
]

type NonSystemMessage = Annotated[
    UserMessage | AssistantMessage | ToolMessage,
    Field(discriminator="role"),
]

type Prompt = Sequence[Message]


class ResponseFormatText(BaseModel):
    type: Literal["text"] = "text"


class ResponseFormatJson(BaseModel):
    type: Literal["json"] = "json"
    json_schema: JsonValue | None = Field(
        default=None,
        validation_alias=AliasChoices("schema", "json_schema"),
        serialization_alias="schema",
    )
    name: str | None = None
    description: str | None = None

    @staticmethod
    def from_model[T: ObjectType](return_type: Type[T]) -> "ResponseFormatJson":
        class Result(_Result[return_type]): ...

        if inspect.isclass(return_type) and issubclass(return_type, BaseModel):
            response_format = return_type
        else:
            response_format = Result
        return ResponseFormatJson(
            json_schema=to_strict_json_schema(response_format),
            name=return_type.__name__,
            description=f"JSON object matching the schema of {return_type.__name__}",
        )


type ResponseFormat = Annotated[
    ResponseFormatText | ResponseFormatJson,
    Field(discriminator="type"),
]



import os
import re
from typing import (
    TYPE_CHECKING,
    AsyncGenerator,
    Literal,
    Optional,
    Sequence,
    TypedDict,
    overload,
)
from pydantic import AnyUrl

from agentia import spec
from agentia.llm.completion import ChatCompletion
from agentia.llm.stream import ChatCompletionEvents, ChatCompletionStream
from agentia.spec.prompt import ObjectType
from agentia.tools.tools import ToolSet, Tool

if TYPE_CHECKING:
    from agentia.agent import Agent
    from agentia.llm.providers import Provider


class GenerationOptions(TypedDict, total=False):
    max_output_tokens: int | None
    temperature: float | None
    stop_sequences: Sequence[str] | None
    top_p: float | None
    top_k: int | None
    presence_penalty: float | None
    frequency_penalty: float | None
    response_format: spec.ResponseFormat | None
    seed: int | None
    tool_choice: spec.ToolChoice | None
    provider_options: spec.ProviderOptions | None
    tools: Sequence[Tool] | ToolSet | None


def get_provider(selector: str) -> "Provider":
    DEFAULT_PROVIDER = os.environ.get("AGENTIA_PROVIDER", "openrouter")
    # if has no scheme, assume gateway
    if re.match(r"^\w+:", selector) is None:
        provider = DEFAULT_PROVIDER
        model = selector
    else:
        uri = AnyUrl(selector)
        provider = uri.scheme
        model = (uri.host or "") + (uri.path or "")
    model = model.strip("/")
    match provider:
        case "openai":
            from .providers.openai import OpenAI

            return OpenAI(model=model)
        case "gateway":
            from .providers.gateway import Gateway

            return Gateway(model=model)
        case "openrouter":
            from .providers.openrouter import OpenRouter

            return OpenRouter(model=model)
        case _:
            raise ValueError(f"Unknown provider: {provider}")


class LLM:
    def __init__(self, model: str) -> None:
        self._provider = get_provider(model)
        self._provider.llm = self
        self._agent: Optional["Agent"] = None

    def __prepare_messages(
        self, prompt: str | spec.Message | Sequence[spec.Message]
    ) -> list[spec.Message]:
        messages: list[spec.Message] = []
        if isinstance(prompt, str):
            messages.append(
                spec.UserMessage(content=[spec.MessagePartText(text=prompt)])
            )
        elif isinstance(prompt, (list, Sequence)):
            messages.extend(prompt)
        else:
            messages.append(prompt)
        return messages

    async def __process_tool_calls(
        self, tool_calls: list[spec.ToolCall], tools: ToolSet | None = None
    ) -> tuple[spec.ToolMessage, list[spec.ToolResult], spec.Message | None]:
        assert tools is not None, "No tools provided"
        tm, tr, fr = await tools.run(self, tool_calls)
        if fr:
            msg = spec.UserMessage(
                content=[
                    spec.MessagePartText(text="[[TOOL_OUTPUT_FILES]]"),
                    *(
                        spec.MessagePartFile(
                            filename=f.id,
                            media_type=f.media_type,
                            data=f.data,
                        )
                        for f in fr
                    ),
                ]
            )
        else:
            msg = None
        return tm, tr, msg

    async def __init_tools(self, options: GenerationOptions | None) -> ToolSet:
        if options is None or "tools" not in options or options["tools"] is None:
            tools = ToolSet([])
        elif isinstance(options["tools"], ToolSet):
            tools = options["tools"]
        else:
            tools = ToolSet(options["tools"])
        await tools.init(self, self._agent)
        return tools

    async def generate_object[T: ObjectType](
        self,
        prompt: str | spec.Message | Sequence[spec.Message],
        type: type[T],
        options: GenerationOptions | None = None,
    ) -> T:
        msgs = await self._generate_object_impl(prompt, type, options)
        assert isinstance(msgs[-1], spec.AssistantMessage)
        return msgs[-1].parse(type)

    async def _generate_object_impl[T: ObjectType](
        self,
        prompt: str | spec.Message | Sequence[spec.Message],
        return_type: type[T],
        options: GenerationOptions | None = None,
    ) -> list[spec.Message]:
        options = options or {}
        options["response_format"] = spec.ResponseFormatJson.from_model(return_type)
        r = self.generate(prompt, options=options)
        await r
        return r.new_messages

    def generate(
        self,
        prompt: str | spec.Message | Sequence[spec.Message],
        options: GenerationOptions | None = None,
    ) -> ChatCompletion:
        options = options or {}

        async def gen() -> (
            AsyncGenerator[spec.AssistantMessage | spec.ToolMessage, None]
        ):
            tools = await self.__init_tools(options)
            messages = self.__prepare_messages(prompt)
            while True:
                result = await self._provider.do_generate(
                    prompt=messages, tool_set=tools, options=options
                )
                c.warnings.extend(result.warnings)
                c.usage += result.usage
                c.new_messages.append(result.message)
                yield result.message
                # Add new messages
                messages.append(result.message)
                tool_calls = result.tool_calls
                if result.finish_reason != "tool-calls":
                    break
                # Call tools and continue
                tool_msg, _, extra_msg = await self.__process_tool_calls(
                    tool_calls, tools=tools
                )
                yield tool_msg
                messages.append(tool_msg)
                c.new_messages.append(messages[-1])
                if extra_msg:
                    messages.append(extra_msg)
                    c.new_messages.append(messages[-1])

        c = ChatCompletion(gen())
        return c

    @overload
    def stream(
        self,
        prompt: str | spec.Message | Sequence[spec.Message],
        /,
        events: Literal[False] = False,
        options: GenerationOptions | None = None,
    ) -> ChatCompletionStream: ...

    @overload
    def stream(
        self,
        prompt: str | spec.Message | Sequence[spec.Message],
        /,
        events: Literal[True],
        options: GenerationOptions | None = None,
    ) -> ChatCompletionEvents: ...

    def stream(
        self,
        prompt: str | spec.Message | Sequence[spec.Message],
        /,
        events: bool = False,
        options: GenerationOptions | None = None,
    ) -> ChatCompletionStream | ChatCompletionEvents:
        options = options or {}

        async def gen() -> AsyncGenerator[spec.StreamPart, None]:
            tools = await self.__init_tools(options)
            messages = self.__prepare_messages(prompt)
            last_finish_reason: spec.FinishReason = "unknown"

            while True:
                tool_calls: list[spec.ToolCall] = []
                parts: list[spec.MessagePart] = []
                last_msg = ""
                last_reasoning = ""
                async for part in self._provider.do_stream(messages, tools, options):

                    match part.type:
                        case "stream-start":
                            s.warnings.extend(part.warnings)
                        case "text-start":
                            last_msg = ""
                        case "text-delta":
                            last_msg += part.delta
                        case "text-end":
                            parts.append(spec.MessagePartText(text=last_msg))
                        case "reasoning-start":
                            last_reasoning = ""
                        case "reasoning-delta":
                            last_reasoning += part.delta
                        case "reasoning-end":
                            parts.append(spec.MessagePartReasoning(text=last_reasoning))
                        case "file":
                            parts.append(
                                spec.MessagePartFile(
                                    data=part.data, media_type=part.media_type
                                )
                            )
                        case "tool-call":
                            parts.append(
                                spec.MessagePartToolCall(
                                    tool_call_id=part.tool_call_id,
                                    tool_name=part.tool_name,
                                    input=part.input,
                                    provider_executed=part.provider_executed,
                                )
                            )
                            if not part.provider_executed:
                                tool_calls.append(part)

                    if part.type == "finish":
                        s.usage += part.usage
                        last_finish_reason = part.finish_reason
                    else:
                        yield part
                messages.append(spec.AssistantMessage(content=parts))
                s.new_messages.append(messages[-1])
                if not tool_calls:
                    break
                # Call tools and continue
                tm, trs, extra_msg = await self.__process_tool_calls(
                    tool_calls, tools=tools
                )
                for tr in trs:
                    yield tr
                messages.append(tm)
                s.new_messages.append(messages[-1])
                if extra_msg:
                    messages.append(extra_msg)
                    s.new_messages.append(messages[-1])
            s.finish_reason = last_finish_reason
            yield spec.StreamPartFinish(usage=s.usage, finish_reason=last_finish_reason)

        if events:
            s = ChatCompletionEvents(gen())
        else:
            s = ChatCompletionStream(gen())
        return s



import inspect
import json
import os
from typing import TYPE_CHECKING, Any, Callable, Optional, Sequence, Union

from pydantic import AliasChoices, BaseModel, Field, JsonValue
from agentia.tools.mcp import MCPServer
import agentia.spec as spec
from agentia.utils.decorators import ToolFuncParam
from inspect import Parameter
from uuid import uuid4

NAME_TAG = "agentia_tool_name"
DISPLAY_NAME_TAG = "agentia_tool_display_name"
IS_TOOL_TAG = "agentia_tool_is_tool"
DESCRIPTION_TAG = "agentia_tool_description"
METADATA_TAG = "agentia_tool_metadata"

if TYPE_CHECKING:
    from ..llm import LLM
    from agentia.plugins import Plugin
    from agentia.agent import Agent


class _BaseTool:
    def __init__(
        self,
        name: str,
        schema: Any,
        display_name: str | None = None,
        description: str | None = None,
        metadata: Any | None = None,
    ):
        self.name = name
        self.display_name = display_name
        self.description = description
        self.schema = schema
        self.metadata = metadata


class _PythonFunctionTool(_BaseTool):
    def __init__(self, f: Callable[..., Any], plugin: Optional["Plugin"] = None):
        if plugin:
            name = getattr(f, NAME_TAG, None) or f"{plugin.name()}__{f.__name__}"
            display_name: str = (
                getattr(f, DISPLAY_NAME_TAG, None) or f"{plugin.name()}@{f.__name__}"
            )
        else:
            name = getattr(f, NAME_TAG, f.__name__)
            display_name = getattr(f, DISPLAY_NAME_TAG, name)
        description: str = getattr(f, DESCRIPTION_TAG, f.__doc__) or ""
        metadata = getattr(f, METADATA_TAG, None)
        params = [
            ToolFuncParam(p, f.__name__)
            for p in inspect.signature(f).parameters.values()
            if p.name != "self"
        ]
        schema = _PythonFunctionTool.get_json_schema(
            name=name, description=description, fparams=params
        )
        super().__init__(
            name=name,
            schema=schema,
            display_name=display_name,
            description=description,
            metadata=metadata,
        )
        self.func = f
        self.plugin = plugin
        self.params = params

    @staticmethod
    def get_json_schema(
        name: str, description: str | None, fparams: list[ToolFuncParam]
    ) -> Any:
        params: Any = {"type": "object", "properties": {}, "required": []}
        for p in fparams:
            if p.required:
                params["required"].append(p.name)
            params["properties"][p.name] = p.schema
        return {
            "name": name,
            "description": description or "",
            "parameters": params,
        }

    def process_json_args(self, llm: "LLM", args: Any):
        from agentia.agent import Agent
        from ..llm import LLM

        p_args: list[Any] = []
        kw_args: dict[str, Any] = {}

        def get_value(name: str, default: Any, is_model: bool):
            if name not in args:
                return default
            if is_model:
                return p.type(**args[name])
            else:
                return args[name]

        assert self.params is not None
        for p in self.params:
            is_model = inspect.isclass(p.type) and issubclass(p.type, BaseModel)
            match p.param.kind:
                case Parameter.POSITIONAL_ONLY if p.param.annotation == Agent:
                    if llm._agent is not None:
                        p_args.append(llm._agent)
                case Parameter.POSITIONAL_ONLY if p.param.annotation == LLM:
                    p_args.append(llm)
                case Parameter.POSITIONAL_ONLY:
                    default = p.default if p.default != Parameter.empty else None
                    p_args.append(get_value(p.name, default, is_model))
                case Parameter.POSITIONAL_OR_KEYWORD | Parameter.KEYWORD_ONLY if (
                    p.param.annotation == Agent
                ):
                    if llm._agent is not None:
                        kw_args[p.name] = llm._agent
                case Parameter.POSITIONAL_OR_KEYWORD | Parameter.KEYWORD_ONLY:
                    default = p.default if p.default != Parameter.empty else None
                    kw_args[p.name] = get_value(p.name, default, is_model)
                case other:
                    raise ValueError(f"{other} is not supported")
        return p_args, kw_args


class _MCPTool(_BaseTool):
    def __init__(self, name: str, schema: Any, server: MCPServer):
        super().__init__(name=name, schema=schema)
        self.server = server


class ProviderTool(BaseModel):
    name: str
    args: dict[str, JsonValue] | None = None


type Tool = Plugin | Callable[..., Any] | ProviderTool | MCPServer
type Tools = Sequence[Tool]


class FileResult(BaseModel):
    id: str | None = None
    data: spec.DataContent
    media_type: str = Field(
        validation_alias=AliasChoices("mediaType", "media_type"),
        serialization_alias="mediaType",
    )


class ToolSet:
    def __init__(self, tools: Tools):
        from agentia.plugins import Plugin

        self.all_tools = tools
        self.plugins: dict[str, Plugin] = {}
        self.mcp_servers: dict[str, MCPServer] = {}
        self.provider_tools: dict[str, ProviderTool] = {}
        self.__tools: dict[str, _BaseTool] = {}
        for t in tools:
            if isinstance(t, Plugin):
                self.__add_plugin(t)
            elif isinstance(t, MCPServer):
                self.__add_mcp_server(t)
            elif isinstance(t, ProviderTool):
                self.provider_tools[t.name] = t
            else:
                assert inspect.isfunction(
                    t
                ), "Expected a function, MCPServer, Plugin, or ProviderTool"
                self.__add_function(t)
        self.__initialized = False

    async def init(self, llm: "LLM", agent: Optional["Agent"] = None):
        if self.__initialized:
            return
        self.__initialized = True
        for plugin in self.plugins.values():
            try:
                plugin.llm = llm
                plugin.agent = agent
                await plugin.init()
            except Exception as e:
                from . import PluginInitError

                raise PluginInitError(plugin.id(), e) from e
        for server in self.mcp_servers.values():
            await server.init()
            tools = server.get_tools()
            for tool in tools:
                self.__tools[tool.name] = tool

    def add(self, t: Union[Callable[..., Any], MCPServer, "Plugin"]) -> None:
        from agentia.plugins import Plugin

        if isinstance(t, Plugin):
            self.__add_plugin(t)
        elif isinstance(t, MCPServer):
            self.__add_mcp_server(t)
        else:
            assert inspect.isfunction(t), "Expected a function, MCPServer, or Plugin"
            self.__add_function(t)

    def __add_function(self, f: Callable[..., Any], plugin: Optional["Plugin"] = None):
        t = _PythonFunctionTool(f=f, plugin=plugin)
        self.__tools[t.name] = t

    def __add_plugin(self, p: "Plugin"):
        # Add all functions from the plugin
        for _n, method in inspect.getmembers(p, predicate=inspect.ismethod):
            if not getattr(method, IS_TOOL_TAG, False):
                continue
            self.__add_function(method, plugin=p)
        # Add the plugin to the list of plugins
        self.plugins[p.id()] = p

    def __add_mcp_server(self, server: MCPServer):
        self.mcp_servers[server.name] = server

    def get_plugin(self, type: type["Plugin"]) -> Optional["Plugin"]:
        for p in self.plugins.values():
            if isinstance(p, type):
                return p
        return None

    def is_empty(self) -> bool:
        return len(self.__tools) == 0

    def get_schema(self) -> Sequence[spec.ProviderDefinedTool | spec.FunctionTool]:
        functions = [
            spec.FunctionTool(
                name=v.name,
                description=v.description or "",
                input_schema=v.schema["parameters"],
            )
            for (k, v) in self.__tools.items()
        ]
        provider_tools = [
            spec.ProviderDefinedTool(id=t.name, name=t.name, args=t.args or {})
            for t in self.provider_tools.values()
        ]
        return functions + provider_tools

    async def __run_python_tool(
        self,
        id: str,
        tool: _PythonFunctionTool,
        llm: "LLM",
        args: Any,
    ) -> tuple[spec.ToolMessage, spec.ToolResult, FileResult | None]:
        p_args, kw_args = tool.process_json_args(llm, args)
        output = tool.func(*p_args, **kw_args)  # type: ignore
        if inspect.isawaitable(output):
            output = await output
        # if output is a FileResult, transform the result
        file = None
        tool_vision = os.getenv("AGENTIA_EXPERIMENTAL_TOOL_FILES", None)
        if tool_vision and tool_vision != "0" and tool_vision.lower() != "false":
            if isinstance(output, FileResult):
                file = output
                if not file.id:
                    file.id = str(uuid4())
                output: Any = {
                    "file_id": file.id,
                    "media_type": file.media_type,
                    "hint": "The tool outputed a file with the given file_id. The file is attached below.",
                }
                d = file.data
                if isinstance(d, str) and (d.startswith(("http://", "https://"))):
                    output["url"] = d
        else:
            if isinstance(output, FileResult):
                output = output.model_dump()
                if isinstance(output["data"], str) and output["data"].startswith(
                    "data:"
                ):
                    del output["data"]
                output["hint"] = "The tool outputed a file with the given file_id."

        tm = spec.ToolMessage(
            content=[
                spec.MessagePartToolResult(
                    tool_call_id=id,
                    tool_name=tool.name,
                    output=spec.ToolResultOutputJson(value=output),
                )
            ]
        )
        tr = spec.ToolResult(
            tool_call_id=id,
            tool_name=tool.name,
            result=output,
        )
        return tm, tr, file

    async def __run_mcp_tool(
        self, id: str, tool: _MCPTool, args: Any
    ) -> tuple[spec.ToolMessage, spec.ToolResult]:
        result = await tool.server.run(tool.name, args)
        tm = spec.ToolMessage(
            content=[
                spec.MessagePartToolResult(
                    tool_call_id=id,
                    tool_name=tool.name,
                    output=spec.ToolResultOutputJson(value=result),
                )
            ]
        )
        tr = spec.ToolResult(
            tool_call_id=id,
            tool_name=tool.name,
            result=result,
        )
        return tm, tr

    async def run(
        self, llm: "LLM", tool_calls: list[spec.ToolCall]
    ) -> tuple[spec.ToolMessage, list[spec.ToolResult], list[FileResult]]:
        tool_results: list[spec.MessagePartToolResult] = []
        tool_results2: list[spec.ToolResult] = []
        file_results: list[FileResult] = []
        for c in tool_calls:
            tool = self.__tools.get(c.tool_name, None)
            if not tool:
                raise ValueError(f"Tool {c.tool_name} not found")
            try:
                args = json.loads(c.input)
                if isinstance(tool, _PythonFunctionTool):
                    tm, tr, file = await self.__run_python_tool(
                        id=c.tool_call_id,
                        tool=tool,
                        llm=llm,
                        args=args,
                    )
                    tool_results.extend(tm.content)
                    tool_results2.append(tr)
                    if file:
                        file_results.append(file)

                elif isinstance(tool, _MCPTool):
                    tm, tr = await self.__run_mcp_tool(
                        id=c.tool_call_id,
                        tool=tool,
                        args=args,
                    )
                    tool_results.extend(tm.content)
                    tool_results2.append(tr)
                else:
                    raise ValueError(f"Tool {c.tool_name} is of unknown type")
            except Exception as e:
                output: JsonValue = {"error": str(e)}
                tool_results.append(
                    spec.MessagePartToolResult(
                        tool_call_id=c.tool_call_id,
                        tool_name=c.tool_name,
                        output=spec.ToolResultOutputErrorJson(value=output),
                    )
                )
        return spec.ToolMessage(content=tool_results), tool_results2, file_results



from typing import TYPE_CHECKING, Annotated, Any, Literal
import tomllib
from pathlib import Path
import os
import importlib.util

from agentia.agent import Agent
from pydantic import AfterValidator, BaseModel, Field, ValidationError

from agentia.tools.tools import ProviderTool
from agentia.tools.mcp import MCPServer

if TYPE_CHECKING:
    from agentia.tools.tools import Tool
    from agentia.plugins import Plugin

DEFAULT_AGENT_CONFIG_PATH = Path.cwd() / "agents"
DEFAULT_AGENT_USER_PLUGIN_PATH = Path.cwd() / "plugins"


class AgentConfig(BaseModel):
    name: str
    icon: str | None = None
    description: str | None = None
    instructions: str | None = None
    user: str | None = None
    model: str | None = None
    plugins: list[str] | None = None
    mcp: list[str] | None = None

    def build_instructions(self) -> str | None:
        if self.name and not self.description:
            return f"YOUR ARE {self.name}\n\n{self.instructions or ''}"
        elif self.description and not self.name:
            return f"YOUR DESCRIPTION: {self.description}\n\n{self.instructions or ''}"
        elif self.description and self.name:
            return (
                f"YOUR ARE {self.name}, {self.description}\n\n{self.instructions or ''}"
            )
        if self.user:
            return f"{self.instructions or ''}\n\nYOU ARE TALKING TO THE USER.\nMY INFO (THE USER): {self.user}"
        else:
            return self.instructions


PluginConfigs = dict[str, dict[str, Any]]


def check_plugins(configs: PluginConfigs) -> PluginConfigs:
    from agentia.plugins import ALL_PLUGINS

    for name, config in configs.items():
        if config is False or config is None:
            continue
        if name not in ALL_PLUGINS:
            raise ValueError(
                f"Unknown plugin: {name}. Available plugins: {', '.join(ALL_PLUGINS.keys())}"
            )
        config = config if isinstance(config, dict) else {}
        ALL_PLUGINS[name].validate_config(config)
    return configs


class MCPLocalConfig(BaseModel):
    command: str
    args: list[str]
    env: dict[str, str] | None = None
    cwd: str | None = None
    type: Literal["local"] = "local"


class MCPHTTPConfig(BaseModel):
    url: str
    headers: dict[str, str] | None = None
    type: Literal["http"] = "http"


class MCPSSEConfig(BaseModel):
    url: str
    headers: dict[str, str] | None = None
    type: Literal["sse"] = "sse"


class MCPWebSocketConfig(BaseModel):
    name: str
    url: str
    type: Literal["websocket"] = "websocket"


MCPConfig = Annotated[
    MCPLocalConfig | MCPHTTPConfig | MCPSSEConfig | MCPWebSocketConfig,
    Field(union_mode="left_to_right"),
]


class Config(BaseModel):
    agent: AgentConfig
    plugins: Annotated[PluginConfigs, AfterValidator(check_plugins)] = Field(
        default_factory=dict
    )
    mcp: dict[str, MCPConfig] = Field(default_factory=dict)
    provider_tools: dict[str, dict[str, Any]] = Field(default_factory=dict)

    def get_enabled_plugins(self) -> list[str]:
        if self.agent.plugins is not None:
            return sorted(self.agent.plugins)
        else:
            return sorted(self.plugins.keys())

    def get_enabled_mcp_servers(self) -> dict[str, MCPConfig]:
        if self.agent.mcp is not None:
            return {name: self.mcp[name] for name in self.agent.mcp if name in self.mcp}
        else:
            return self.mcp


class AgentInfo(BaseModel):
    id: str
    config_path: Path
    config: "Config"


def __create_provider_tools(config: Config) -> list[ProviderTool]:
    ptools: list[ProviderTool] = []
    for name, args in config.provider_tools.items():
        ptools.append(ProviderTool(name=name, args=args))
    return ptools


def __create_plugins(config: Config) -> tuple[list["Plugin"], dict[str, Any]]:
    from agentia.plugins import ALL_PLUGINS

    plugins: list["Plugin"] = []
    plugin_configs = {}
    enabled_plugins = config.get_enabled_plugins()
    for name in enabled_plugins:
        c = config.plugins.get(name, {})
        if name not in ALL_PLUGINS:
            raise ValueError(f"Unknown plugin: {name}")
        PluginCls = ALL_PLUGINS[name]
        if not (c is None or isinstance(c, dict)):
            raise ValueError(
                f"Invalid config for plugin {name}: must be a dict or null"
            )
        c = c if isinstance(c, dict) else {}
        plugin_configs[name] = c
        plugins.append(PluginCls.instantiate(config=c))
    return plugins, plugin_configs


def __create_mcp_servers(config: Config) -> dict[str, "MCPServer"]:
    mcps: dict[str, MCPServer] = {}
    for name, cfg in config.mcp.items():
        match cfg.type:
            case "local":
                mcps[name] = MCPServer(
                    name=name,
                    command=cfg.command,
                    args=cfg.args,
                    env=cfg.env,
                    cwd=cfg.cwd,
                )
            case "http":
                mcps[name] = MCPServer(
                    name=name,
                    type="http",
                    url=cfg.url,
                    headers=cfg.headers,
                )
            case "sse":
                mcps[name] = MCPServer(
                    name=name,
                    type="sse",
                    url=cfg.url,
                    headers=cfg.headers,
                )
            case "websocket":
                mcps[name] = MCPServer(
                    name=name,
                    type="websocket",
                    url=cfg.url,
                )
            case _:
                raise ValueError(f"Unknown MCP server type: {cfg.type}")
    return mcps


def __load_agent_from_config(file_or_config: Path | Config) -> Agent:
    """Load a bot from a configuration file"""

    # Load the configuration file
    if isinstance(file_or_config, Config):
        config = file_or_config
    else:
        assert file_or_config.exists()
        file = file_or_config.resolve()
        try:
            config = Config(**tomllib.loads(file.read_text()))
        except ValidationError as e:
            raise ValueError(f"Invalid config file: {file}\n{repr(e)}") from e
    # Create tools
    plugins, _ = __create_plugins(config)
    mcp_servers = __create_mcp_servers(config)
    provider_tools = __create_provider_tools(config)
    tools: list["Tool"] = [*plugins, *mcp_servers.values(), *provider_tools]
    # Create agent
    agent_id = file_or_config.stem if isinstance(file_or_config, Path) else None
    agent = Agent(
        id=agent_id,
        name=config.agent.name,
        icon=config.agent.icon,
        description=config.agent.description,
        model=config.agent.model,
        tools=tools,
        instructions=config.agent.build_instructions(),
    )
    return agent


__user_plugins_loaded = False


def prepare_user_plugins():
    from agentia import LOGGER

    global __user_plugins_loaded
    if __user_plugins_loaded:
        return
    __user_plugins_loaded = True
    if path := os.environ.get("AGENTIA_USER_PLUGIN_DIR"):
        plugins_dir = Path(path)
    else:
        plugins_dir = DEFAULT_AGENT_USER_PLUGIN_PATH
    if not plugins_dir.is_dir():
        return
    LOGGER.info(f"Loading user plugins from {plugins_dir}")
    # Install requirements.txt
    if (plugins_dir / "requirements.txt").is_file():
        ret = os.system(f"uv pip install -r {plugins_dir / "requirements.txt"}")
        if ret != 0:
            raise RuntimeError(
                f"Failed to install requirements.txt from {plugins_dir}: {ret}"
            )
    # Load plugins
    for file in plugins_dir.glob("*.py"):
        if not file.is_file() or file.stem.startswith(("_", ".", "-")):
            continue
        name = file.stem
        LOGGER.info(f" - {name} ({file})")
        spec = importlib.util.spec_from_file_location(name, file)
        assert spec is not None
        assert spec.loader is not None
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)


def load_agent_from_config(config_path: str | Path | Config) -> Agent:
    """Load a bot from a configuration file"""
    if isinstance(config_path, Config):
        return __load_agent_from_config(config_path)
    config_path = Path(config_path) if isinstance(config_path, str) else config_path
    assert config_path.suffix == ".toml", "Agent config file must be a .toml file"
    config_path = config_path.resolve()
    return __load_agent_from_config(config_path)



import base64
from collections.abc import AsyncGenerator
import json
import os
from typing import Any, override
from uuid import uuid4

from agentia.tools.tools import ToolSet
from . import GenerationOptions, ProviderGenerationResult, Provider

from ...spec import *
from openai.types.chat import (
    ChatCompletionMessageParam,
    ChatCompletionToolMessageParam,
    ChatCompletionUserMessageParam,
    ChatCompletionSystemMessageParam,
    ChatCompletionAssistantMessageParam,
    ChatCompletionMessageToolCallParam,
    ChatCompletionContentPartParam,
    ChatCompletionContentPartTextParam,
    ChatCompletionContentPartImageParam,
    ChatCompletionContentPartInputAudioParam,
    ChatCompletionToolParam,
    ChatCompletionToolChoiceOptionParam,
)
import openai
from openai.types import CompletionUsage


def _gen_id() -> str:
    return uuid4().hex


class OpenAIAPIProvider(Provider):
    def __init__(
        self,
        provider: str,
        model: str,
        api_key: str | None = None,
        base_url: str | None = None,
    ):
        if model.endswith(":think"):
            model = model[: -len(":think")]
            think = True
        else:
            think = False
        super().__init__(provider=provider, model=model)
        api_key = api_key or os.environ.get("OPENAI_API_KEY")
        if not api_key:
            raise ValueError("OPENAI_API_KEY environment variable not set")
        self.client = openai.AsyncOpenAI(api_key=api_key, base_url=base_url)
        self.extra_headers: dict[str, str] = {}
        self.extra_body: dict[str, Any] = {
            "modalities": ["text", "image"],
        }
        if think:
            self.reasoning = True
            self.enable_reasoning()
        else:
            self.reasoning = False

    def enable_reasoning(self) -> None:
        self.extra_body["reasoning"] = {
            "enabled": True,
            "effort": "high",
        }

    def _to_oai_content_part(
        self, p: MessagePart, index: int
    ) -> ChatCompletionContentPartParam:
        match p.type:
            case "text":
                return ChatCompletionContentPartTextParam(type="text", text=p.text)
            case "file":
                if p.media_type.startswith("image/"):
                    if isinstance(p.data, str):
                        if p.data.startswith("data:"):
                            # this is a data URL
                            url = p.data
                        elif p.data.startswith("http://") or p.data.startswith(
                            "https://"
                        ):
                            # this is a URL
                            url = p.data
                        else:
                            # this is a base64 encoded string
                            url = f"data:{p.media_type};base64,{p.data}"
                    elif isinstance(p.data, HttpUrl):
                        url = str(p.data)
                    else:
                        assert isinstance(p.data, bytes)
                        base64_data = base64.b64encode(p.data).decode(encoding="utf-8")
                        url = f"data:{p.media_type};base64,{base64_data}"
                    detail = (p.provider_options or {}).get("imageDetail", None)
                    if detail not in ["auto", "low", "high"]:
                        detail = "auto"
                    return ChatCompletionContentPartImageParam(
                        type="image_url",
                        image_url={"url": url, "detail": detail},  # type: ignore
                    )
                elif p.media_type.startswith("audio/"):
                    if isinstance(p.data, HttpUrl) or (
                        isinstance(p.data, str)
                        and (
                            p.data.startswith("http://")
                            or p.data.startswith("https://")
                        )
                    ):
                        raise ValueError("audio file parts with URLs")
                    if isinstance(p.data, str):
                        data = p.data
                    else:
                        assert isinstance(p.data, bytes)
                        data = p.data.decode(encoding="utf-8")
                    format: Literal["wav", "mp3"]
                    if p.media_type == "audio/wav":
                        format = "wav"
                    elif p.media_type == "audio/mpeg" or p.media_type == "audio/mp3":
                        format = "mp3"
                    else:
                        raise ValueError(
                            f"audio file parts with media type {p.media_type}"
                        )
                    return ChatCompletionContentPartInputAudioParam(
                        type="input_audio", input_audio={"data": data, "format": format}
                    )
                else:
                    if isinstance(p.data, str) and p.data.startswith("file-"):
                        # this is a file ID
                        return {"type": "file", "file": {"file_id": p.data}}
                    else:
                        url = None
                        if isinstance(p.data, str):
                            if p.data.startswith("data:"):
                                url = p.data
                            else:
                                # this is a base64 encoded string
                                url = f"data:{p.media_type};base64,{p.data}"
                        elif isinstance(p.data, HttpUrl):
                            url = str(p.data)
                        else:
                            assert isinstance(p.data, bytes)
                            base64_data = base64.b64encode(p.data).decode(
                                encoding="utf-8"
                            )
                            url = f"data:{p.media_type};base64,{base64_data}"
                        return {
                            "type": "file",
                            "file": {
                                "filename": p.filename or f"file-{index}",
                                "file_data": url,
                            },
                        }
            case _:
                raise ValueError(f"content part of type {p.type}")

    def _to_oai_messages(self, prompt: Prompt) -> list[ChatCompletionMessageParam]:
        r: list[ChatCompletionMessageParam] = []
        for m in prompt:
            if m.role == "system":
                r.append(
                    ChatCompletionSystemMessageParam(role="system", content=m.content)
                )
            elif m.role == "user":
                if len(m.content) == 1 and m.content[0].type == "text":
                    r.append(
                        ChatCompletionUserMessageParam(
                            role="user", content=m.content[0].text
                        )
                    )
                    continue
                parts: list[ChatCompletionContentPartParam] = []
                for i, p in enumerate(m.content):
                    parts.append(self._to_oai_content_part(p, i))
                r.append(ChatCompletionUserMessageParam(role="user", content=parts))
            elif m.role == "assistant":
                text = ""
                tool_calls: list[ChatCompletionMessageToolCallParam] = []
                for i, p in enumerate(m.content):
                    if p.type == "text":
                        text += p.text
                    elif p.type == "tool-call":
                        tool_calls.append(
                            {
                                "id": p.tool_call_id,
                                "type": "function",
                                "function": {
                                    "name": p.tool_name,
                                    "arguments": json.dumps(p.input),
                                },
                            }
                        )
                r.append(
                    ChatCompletionAssistantMessageParam(
                        role="assistant",
                        content=text if text else None,
                        tool_calls=tool_calls,
                    )
                )
            elif m.role == "tool":
                for p in m.content:
                    out = p.output
                    if out.type == "text" or out.type == "error-text":
                        val = out.value
                    elif out.type == "execution-denied":
                        val = out.reason or "Tool execution denied"
                    else:
                        val = json.dumps(out.value)

                    r.append(
                        ChatCompletionToolMessageParam(
                            role="tool", tool_call_id=p.tool_call_id, content=val
                        )
                    )
            else:
                raise ValueError(f"Unsupported message role {m.role}")
        return r

    def _get_finish_reason(
        self,
        choice: Literal[
            "stop", "length", "tool_calls", "content_filter", "function_call"
        ],
    ) -> FinishReason:
        match choice:
            case "stop":
                return "stop"
            case "length":
                return "length"
            case "tool_calls" | "function_call":
                return "tool-calls"
            case "content_filter":
                return "content-filter"
            case _:
                return "unknown"

    def _prepare_tools(
        self,
        tools: Sequence[Tool] | None,
        tool_choice: ToolChoice | None,
    ) -> tuple[
        list[ChatCompletionToolParam] | None,
        ChatCompletionToolChoiceOptionParam | None,
        list[Warning],
    ]:
        oai_tools: list[ChatCompletionToolParam] = []
        warnings: list[Warning] = []
        if not tools:
            return None, None, []
        for t in tools:
            match t.type:
                case "function":
                    f = {
                        "name": t.name,
                        "description": t.description,
                        "parameters": t.input_schema,  # type: ignore
                    }
                    for k in list(f.keys()):
                        if f[k] is None:
                            del f[k]
                    oai_tools.append(
                        {
                            "type": "function",
                            "function": f,  # type: ignore
                        }
                    )
                case _:
                    warnings.append(UnsupportedToolWarning(tool=t))
        if not tool_choice:
            return oai_tools, None, warnings
        match tool_choice.type:
            case "none":
                return oai_tools, "none", warnings
            case "auto":
                return oai_tools, "auto", warnings
            case "required":
                return oai_tools, "required", warnings
            case "tool":
                tc: ChatCompletionToolChoiceOptionParam = {
                    "type": "function",
                    "function": {"name": tool_choice.tool_name},
                }
                return oai_tools, tc, warnings

    def _get_usage(self, u: CompletionUsage | None) -> Usage:
        if not u:
            return Usage()
        return Usage(
            input_tokens=u.prompt_tokens,
            output_tokens=u.completion_tokens,
            total_tokens=u.total_tokens,
            reasoning_tokens=(
                u.completion_tokens_details.reasoning_tokens
                if u.completion_tokens_details
                else None
            ),
            cached_input_tokens=(
                u.prompt_tokens_details.cached_tokens
                if u.prompt_tokens_details
                else None
            ),
        )

    def _prepare_args(
        self, prompt: Prompt, tool_set: ToolSet, options: GenerationOptions
    ) -> tuple[dict[str, Any], list[Warning]]:
        msgs = self._to_oai_messages(prompt)
        rf = options.get("response_format", None)
        if not rf or rf.type == "text":
            response_format = None
        elif not rf.json_schema:
            response_format = {"type": "json_object"}
        else:
            response_format = {
                "type": "json_schema",
                "json_schema": {
                    "schema": rf.json_schema,
                    "strict": True,
                    "name": rf.name or "response",
                    "description": rf.description,
                },
            }
        warnings: list[Warning] = []
        schema = tool_set.get_schema()
        tools, tool_choice, tool_warnings = self._prepare_tools(
            schema, options.get("tool_choice", None)
        )
        warnings.extend(tool_warnings)
        args: dict[str, Any] = {
            "model": self.model,
            "messages": msgs,
            "temperature": options.get("temperature", None),
            "top_p": options.get("top_p", None),
            "max_tokens": options.get("max_output_tokens", None),
            "frequency_penalty": options.get("frequency_penalty", None),
            "presence_penalty": options.get("presence_penalty", None),
            "response_format": response_format,
            "stop": options.get("stop_sequences", None),
            "seed": options.get("seed", None),
            "tools": tools,
            "tool_choice": tool_choice,
        }
        for k in list(args.keys()):
            if args[k] is None:
                del args[k]
        return args, warnings

    @override
    async def do_generate(
        self, prompt: Prompt, tool_set: ToolSet, options: GenerationOptions
    ) -> ProviderGenerationResult:
        args, warnings = self._prepare_args(prompt, tool_set, options)
        response = await self.client.chat.completions.create(
            **args,
            extra_headers=self.extra_headers,
            extra_body=self.extra_body,
            stream=False,
        )
        choice = response.choices[0]
        content: Sequence[Content] = []
        tool_calls: list[ToolCall] = []

        if self.reasoning and hasattr(choice.message, "reasoning"):
            reasoning_text: str = choice.message.reasoning  # type: ignore
            if reasoning_text and reasoning_text.strip() != "":
                content.append(Reasoning(text=reasoning_text))

        text = choice.message.content
        if text:
            content.append(Text(type="text", text=text))

        for tc in choice.message.tool_calls or []:
            assert tc.type == "function"
            t = ToolCall(
                type="tool-call",
                tool_call_id=tc.id or _gen_id(),
                tool_name=tc.function.name,
                input=tc.function.arguments,
            )
            content.append(t)
            tool_calls.append(t)

        for a in choice.message.annotations or []:
            content.append(
                SourceURL(
                    type="source",
                    source_type="url",
                    id=_gen_id(),
                    url=a.url_citation.url,
                    title=a.url_citation.title,
                )
            )

        if images := getattr(choice.message, "images", []):
            for i in images:
                if i["type"] == "image_url":
                    url = i["image_url"]["url"]
                    media_type = None
                    if url.startswith("data:"):
                        media_type = url.split(";")[0][5:]
                    elif url.startswith("http://") or url.startswith("https://"):
                        if url.lower().endswith(
                            (".png", ".jpg", ".jpeg", ".gif", ".bmp", ".webp")
                        ):
                            media_type = "image/" + url.split(".")[-1].lower()
                    content.append(File(media_type=media_type or "image/png", data=url))

        return ProviderGenerationResult(
            message=AssistantMessage.from_contents(content),
            tool_calls=tool_calls,
            warnings=warnings,
            finish_reason=self._get_finish_reason(choice.finish_reason),
            usage=self._get_usage(response.usage),
            provider_metadata=None,
        )

    @override
    async def do_stream(
        self, prompt: Prompt, tool_set: ToolSet, options: GenerationOptions
    ) -> AsyncGenerator[StreamPart, None]:
        args, warnings = self._prepare_args(prompt, tool_set, options)
        response = await self.client.chat.completions.create(
            **args,
            extra_headers=self.extra_headers,
            extra_body=self.extra_body,
            stream=True,
            stream_options={
                "include_usage": True,
            },
        )
        started = False
        streaming_text = False
        streaming_reasoning = False
        tool_calls: list[ToolCall | None] = []
        async for chunk in response:
            if not started:
                started = True
                yield StreamPartStreamStart(warnings=[])
                yield StreamPartResponseMetadata(
                    id=_gen_id(),
                    timestamp=datetime.fromtimestamp(chunk.created),
                    model_id=self.model,
                )
            if not chunk.choices:
                yield StreamPartFinish(
                    usage=self._get_usage(chunk.usage), finish_reason="stop"
                )
                continue

            choice = chunk.choices[0]
            delta = choice.delta

            if rd := getattr(delta, "reasoning", None):
                if not self.reasoning:
                    continue
                if not streaming_reasoning:
                    streaming_reasoning = True
                    yield StreamPartReasoningStart(id=_gen_id())
                yield StreamPartReasoningDelta(id=_gen_id(), delta=rd)
            elif streaming_reasoning:
                streaming_reasoning = False
                yield StreamPartReasoningEnd(id=_gen_id())

            if delta.content:
                if not streaming_text:
                    streaming_text = True
                    yield StreamPartTextStart(id=_gen_id())
                yield StreamPartTextDelta(id=_gen_id(), delta=delta.content)
            elif streaming_text:
                streaming_text = False
                yield StreamPartTextEnd(id=_gen_id())

            if delta.tool_calls:
                for tc in delta.tool_calls:
                    index = tc.index
                    if index >= len(tool_calls) or tool_calls[index] is None:
                        tool_calls.extend([None] * (index - len(tool_calls) + 1))
                        assert tc.id and tc.function and tc.function.name
                        yield StreamPartToolInputStart(
                            id=tc.id,
                            tool_name=tc.function.name,
                        )
                        tool_calls[index] = ToolCall(
                            tool_call_id=tc.id,
                            tool_name=tc.function.name,
                            input="",
                        )
                    tc_obj = tool_calls[index]
                    assert tc_obj is not None
                    if tc.function and tc.function.arguments:
                        yield StreamPartToolInputDelta(
                            id=tc_obj.tool_call_id,
                            delta=tc.function.arguments or "",
                        )
                        tc_obj.input += tc.function.arguments or ""

                for i, tc in enumerate(tool_calls):
                    if not tc or not tc.input:
                        continue
                    try:
                        json.loads(tc.input)
                        yield StreamPartToolInputEnd(id=tc.tool_call_id)
                        yield tc
                        tool_calls[i] = None
                    except Exception:
                        continue

            if choice.finish_reason:
                yield StreamPartFinish(
                    usage=self._get_usage(chunk.usage),
                    finish_reason=self._get_finish_reason(choice.finish_reason),
                )
